import numpy as np
from sphere_array import get_A_inc, get_T_comp_t_n, get_C_sum_comp_t, get_E_sca_wj

from math_functions import get_gaussian, get_IFT, get_FT, get_modes, get_E_inc_wj
from sphere_functions import S
import matplotlib.pyplot as plt
import cmath
from time import perf_counter
import string

t_i = perf_counter()

"""
--input
A(matrix): square matrix to be conditioned

--output
A_new: new square matrix with better conditioning
C: column conditioner
"""


"""natural constants in SI units"""

e = 1.6022 * 1e-19  # charge on an electron

m_e = 9.1094 * 1e-31  # mass of an electron

eps_o = 8.8542 * 1e-12  # free space permittivity

c_o = 3 * 1e8  # speed of light in vaccum


"""material parameters"""

c = 1.5 * 1e15  # dimensionless constant

wn = 4 * c  # resonance frequency

ln = 2 * np.pi * c_o / wn

lm = 600 * 1e-9

# print(lm/ln)

R = 1.1 * 2 * np.pi * c_o / wn  # distance parameter

r1 = 0.6 * R  # radius of the sphere

yn = wn / 8  # damping ratio

No = (11 * wn ** 2 * eps_o * m_e) / (e ** 2)  # free electron density


"""time modulation parameters"""

wm = wn / 10  # modulation frequency

Ms = 0.9  # modulation strength


"""incident wave"""

wo = 3.1 * wm  # centering frequency of the pulse

To = 2.9 * 2 * np.pi / wn  # pulse duration

to = 8 * To  # pulse delay

Eo = 1  # time domain amplitude

w_min = wo - (3 / To)

w_max = wo + (20 / To)  # w_min, max chosen such that most of pulse is covered

polpw = 1  # polarisation of incident wave

sigma_f = 1 / To

sigma_t = To

"""floquet comb"""


dw = wm / 50

Omega_f = np.arange(0.0001 * wm, wm, dw)  # Floquet frequencies

jo = int(w_min / wm)  # appropriately chosen integer

NOm = int(w_max / wm - jo) + 1  # total number of frequencies in each floquet comb


"""Lattice parameters"""

a = np.array([[3 * R, 0], [0, 3 * R]])  # lattice vectors

rs = np.zeros(3)  # shift vector


"""Miscellaneous"""

l_max = 3  # maximum order of the multipole considered

ru = np.array(
    [0.75 * R, 0.75 * R, 1.2 * r1]
)  # point at which the scattered field is calculated

rd = np.array([0.75 * R, 0.75 * R, -1.2 * r1])

modes = get_modes(l_max)

"""calculations and output results"""


E_w_scu = np.zeros(0)  # null array for scattered field amplitude of final output

E_w_scd = np.zeros(0)  # null array for scattered field amplitude of final output


w_sc = np.zeros(0)  # null array for scattered frequencies of final output

E_sca_aru = np.zeros(NOm, dtype="complex")  # scattered field array for each w[i]
E_sca_ard = np.zeros(NOm, dtype="complex")  # scattered field array for each w[i]

w_d = np.zeros(NOm)  # domain array for each w_arr[i]

kpar = np.array([0, 0 * wo / c_o])  # parallel wave vector of the incident wave

for i in range(0, len(Omega_f)):
    
    A_inc_comp_t = np.zeros(0)  # null array for each floquet frequency

    k, S_k, k_o_ar = S(Omega_f[i], NOm, jo, No, wn, yn, Ms, wm)

    T_comp_t = get_T_comp_t_n(l_max, NOm, k, S_k, k_o_ar, r1)

    C_sum_comp_t = get_C_sum_comp_t(modes, jo, wm, NOm, Omega_f[i], kpar, a, rs, l_max)

    for j in range(0, NOm):

        wj = Omega_f[i] + (j + jo) * wm

        kj = wj / c_o

        if wj >= 0:

            kz = cmath.sqrt(kj ** 2 - kpar[0] ** 2 - kpar[1] ** 2)

            A_inc = get_gaussian(wj, wo, to, To, Eo) * get_A_inc(
                kj, kpar[0], kpar[1], kz, polpw, l_max
            )

        else:

            kz = -np.conjugate(cmath.sqrt(kj ** 2 - kpar[0] ** 2 - kpar[1] ** 2))

            A_inc = np.conjugate(get_gaussian(-wj, wo, to, To, Eo)) * get_A_inc(
                kj, kpar[0], kpar[1], kz, polpw, l_max
            )

        A_inc_comp_t = np.append(A_inc_comp_t, A_inc)

        A_inc_comp_t = A_inc_comp_t[:, np.newaxis]

    I = np.eye(len(T_comp_t))

    A = I - (T_comp_t @ C_sum_comp_t)

    B = T_comp_t

    Teff = np.linalg.inv(A) @ B

    A_sca = Teff @ A_inc_comp_t

    for j in range(0, NOm):  # field calculation for each floquet comb

        w_d[j] = Omega_f[i] + (j + jo) * wm

        kinc = np.array([0, 0, w_d[j] / c_o])

        E_sca_aru[j] = (
            get_E_sca_wj(kpar, w_d[j], j, l_max, a, ru, A_sca)[0]
            + 0 * get_E_inc_wj(w_d[j], wo, to, To, Eo, polpw, kinc, ru)[0]
        )
        E_sca_ard[j] = (
            get_E_sca_wj(kpar, w_d[j], j, l_max, a, rd, A_sca)[0]
            + 0 * get_E_inc_wj(w_d[j], wo, to, To, Eo, polpw, kinc, rd)[0]
        )

    w_sc = np.append(w_sc, w_d)

    E_w_scu = np.append(E_w_scu, E_sca_aru)
    E_w_scd = np.append(E_w_scd, E_sca_ard)
    print("iterations done:" + str(i + 1) + "/" + str(len(Omega_f)))

print('generating figure...')
###########arranging in ascending order###########

idx = np.argsort(w_sc)

w_sc, E_w_scu = w_sc[idx], E_w_scu[idx]
E_w_scd = E_w_scd[idx]


##interpolation to zero and adding conjugate terms##

wl = np.arange(0 + dw, min(w_sc), dw)

E_wl = np.zeros(len(wl), dtype="complex")

w_sc = np.hstack([-w_sc, -wl, [0], wl, w_sc])

E_w_scu = np.hstack([np.conjugate(E_w_scu), E_wl, [0 + 0j], E_wl, E_w_scu])
E_w_scd = np.hstack([np.conjugate(E_w_scd), E_wl, [0 + 0j], E_wl, E_w_scd])

idx = np.argsort(w_sc)

w_sc, E_w_scu = w_sc[idx], E_w_scu[idx]
E_w_scd = E_w_scd[idx]

####### zero padding in frequency domain #############

w1 = np.arange(20 * min(w_sc), min(w_sc), dw)

w2 = np.arange(max(w_sc), 20 * max(w_sc), dw)

E_w1 = np.zeros(len(w1), dtype="complex")

E_w2 = np.zeros(len(w2), dtype="complex")

w_sc, E_w_scu = np.hstack([w1, w_sc, w2]), np.hstack([E_w1, E_w_scu, E_w2])
E_w_scd = np.hstack([E_w1, E_w_scd, E_w2])


###############time-domain signal #####################

t, E_tu = get_IFT(w_sc, E_w_scu)
t, E_td = get_IFT(w_sc, E_w_scd)


#############zero padding in time-domain###############

dt = t[1] - t[0]

t1 = np.arange(max(t), 200 * max(t), dt)

E_t1 = np.zeros(len(t1), dtype="complex")

t = np.hstack([t, t1])

E_tu = np.hstack([E_tu, E_t1])
E_td = np.hstack([E_td, E_t1])

w_sc, E_w_scu = get_FT(t, E_tu)
w_sc, E_w_scd = get_FT(t, E_td)


###################plotting#########################


"""incident field in time domain"""


def get_E_inc_t(t, wo, to, To, r):

    b = t - to - r[2] / c_o

    E_inc_t = np.exp(-(b ** 2) / (2 * To ** 2)) * np.cos(wo * b)

    return E_inc_t


"""loading comsol file"""

a = np.loadtxt("nnoffaxis_12a.txt", skiprows=5)
b = np.loadtxt("nnoffaxis_12b.txt", skiprows=5)

ru = np.array([0.75 * R, 0.75 * R, 1.2 * r1])
rd = np.array([0.75 * R, 0.75 * R, -1.2 * r1])

Tcu = to + ru[2] / c_o
Tcd = to + rd[2] / c_o

"""time domain data"""

tu = a[:, 0]
td = b[:, 0]

E_xu = a[:, 1]
E_xd = b[:, 1]

E_xu = E_xu - get_E_inc_t(tu, wo, to, To, ru)
E_xd = E_xd - get_E_inc_t(td, wo, to, To, rd)


""" fourier transform"""

wu, E_xwu = get_FT(tu, E_xu)
wd, E_xwd = get_FT(td, E_xd)

E_wu = np.abs(E_xwu)
E_wd = np.abs(E_xwd)

"""plotting"""

FTu = E_w_scu
FTd = E_w_scd
w1 = w_sc
w2 = w_sc
Tu = np.real(E_tu)
Td = np.real(E_td)
t1 = t


fig, ax = plt.subplots(2, 2, figsize=(7, 3))

l1 = ax[0, 0].plot(np.real(w1) / wn, wn * np.abs(FTu) / (2 * np.pi), label="Analytical")
l2 = ax[0, 0].plot(
    wu / wn, wn * E_wu / (2 * np.pi), "--", label="numerical", color="red", linewidth=2
)
ax[0, 0].axvspan(
    (wo - 3 * sigma_f) / wn, (wo + 3 * sigma_f) / wn, color="g", alpha=0.1, lw=0
)
ax[0, 0].set_xlim(0, 1.5)
ax[0, 0].set_ylim(0, 1)
ax[0, 0].set_xticks([0, 0.75, 1.5])
ax[0, 0].set_xticklabels([0, 0.75, 1.5], fontsize=10)
ax[0, 0].set_ylabel(
    r"$\frac{\omega_n}{2\pi}|E_\mathrm{sca}(\omega)\hat{\mathbf{x}}|$", fontsize=10
)
ax[0, 0].grid()


ax[1, 0].plot(np.real(w2) / wn, wn * np.abs(FTd) / (2 * np.pi))
ax[1, 0].plot(wd / wn, wn * E_wd / (2 * np.pi), "--", color="red", linewidth=2)
ax[1, 0].axvspan(
    (wo - 3 * sigma_f) / wn, (wo + 3 * sigma_f) / wn, color="g", alpha=0.1, lw=0
)
ax[1, 0].set_xlim(0, 1.5)
ax[1, 0].set_ylim(0, 1.2)
ax[1, 0].set_xticks([0, 0.75, 1.5])
ax[1, 0].set_xticklabels([0, 0.75, 1.5], fontsize=10)
ax[1, 0].set_ylabel(
    r"$\frac{\omega_n}{2\pi}|E_\mathrm{sca}(\omega)\hat{\mathbf{x}}|$", fontsize=10
)
ax[1, 0].set_xlabel("${\omega}/{\omega_n}$", fontsize=10)
ax[1, 0].grid()

ax[0, 1].plot(
    t1 / To, Tu,
)
ax[0, 1].plot(tu / To, E_xu, "--", color="red")
ax[0, 1].axvspan(
    (Tcu - 3 * sigma_t) / To, (Tcu + 3 * sigma_t) / To, color="g", alpha=0.1, lw=0
)
ax[0, 1].set_xlim(5, 30)
ax[0, 1].set_ylabel("$E_\mathrm{sca}(t)\hat{\mathbf{x}}$", fontsize=10)
ax[0, 1].set_xticks([5, 15, 30])
ax[0, 1].set_xticklabels([5, 15, 30], fontsize=10)
ax[0, 1].set_ylim(-0.6, 0.6)
ax[0, 1].grid()

ax[1, 1].plot(t1 / To, Td)
ax[1, 1].plot(td / To, E_xd, "--", color="red")
ax[1, 1].axvspan(
    (Tcd - 3 * sigma_t) / To, (Tcd + 3 * sigma_t) / To, color="g", alpha=0.1, lw=0
)
ax[1, 1].set_xlim(5, 30)
ax[1, 1].set_ylim(-0.6, 0.6)
ax[1, 1].set_ylabel("$E_\mathrm{sca}(t)\hat{\mathbf{x}}$", fontsize=10)
ax[1, 1].set_xticks([5, 15, 30])
ax[1, 1].set_xticklabels([5, 15, 30], fontsize=10)
ax[1, 1].set_xlabel("${T}/{T_0}$", fontsize=10)
ax[1, 1].grid()
line_labels = ["Analytical", "Numerical"]
fig.legend(
    labels=line_labels,  # The labels for each line
    loc="upper center",  # Position of legend
    borderaxespad=0.05,  # Small spacing around legend box.
    ncol=2,
    fancybox=True,
    fontsize=10,
    # bbox_to_anchor=(-0.5, 1.1)
)
fig.tight_layout()
fig.subplots_adjust(top=0.9)

ax = ax.flat
for n, axs in enumerate(ax):

    axs.text(
        0.9, 0.8, string.ascii_lowercase[n] + ")", transform=axs.transAxes, size=10
    )

print(perf_counter() - t_i)
