#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  2 00:03:33 2022

@author: puneetgarg
"""

import numpy as np
from sphere_array import (
    get_A_inc,
    get_A_inc_comp_t,
    get_T_comp_t_n,
    get_C_sum_comp_t,
    get_Ejpqs_coeff,
)
from sphere_functions import S
from math_functions import get_modes
import matplotlib.pyplot as plt
import cmath
import string


"""natural constants in SI units"""

e = 1.6 * 1e-19  # charge on an electron

m_e = 9.11 * 1e-31  # mass of an electron

eps_o = 8.854 * 1e-12  # free space permittivity

c_o = 3 * 1e8  # speed of light in vaccum


"""material parameters"""

c = 1.5 * 1e15  # dimensionless constant

wn = 4 * c  # resonance frequency

yn = wn / 8

No = (11 * wn ** 2 * eps_o * m_e) / (e ** 2)  # free electron density

R = 216 * 1e-9

"""time modulation parameters"""

wm = wn / 10

Ms = 0.9

"""incident wave"""

winc = np.arange(0.0187 * wn, 1.012 * wn, wn / 1000)

ko = winc / c_o  # free space wavenumber

sinc = 0  # polarisation of incident wave

kpar = np.array([0, 0])  # works only for perpendicular inc

"""floquet comb"""

jo = -4  # appropriately chosen integer

NOm = 8  # total number of frequencies in the floquet comb


"""Lattice parameters"""


a1 = 610

w_l = (2 * np.pi * c_o) / (a1 * 1e-9)

a = np.array([[a1 * 1e-9, 0], [0, a1 * 1e-9]])  # lattice vectors

rs = np.zeros(3)  # shift vector


"""Miscellaneous"""

l_max = 3  # maximum considered order of the multipole

dim = 2 * l_max * (l_max + 2) * NOm

modes = get_modes(l_max)


data_m1 = np.zeros(dim * len(winc), dtype="complex").reshape(dim, len(winc))

# calculating data matrix each column of whose contains Asca for a given winc

for i in range(0, len(winc)):

    """calculations using functions defined before"""

    A_inc = get_A_inc(
        ko[i], kpar[0], kpar[1], winc[i] / c_o, sinc, l_max
    )  # incident field, only works for normal incidence

    A_inc_comp_t = get_A_inc_comp_t(jo, NOm, A_inc)

    k, S_k, k_o_ar = S(winc[i], NOm, jo, No, wn, yn, Ms, wm)

    T_comp_t = get_T_comp_t_n(l_max, NOm, k, S_k, k_o_ar, R)  # of single sphere

    C_sum_comp_t = get_C_sum_comp_t(modes, jo, wm, NOm, winc[i], kpar, a, rs, l_max)

    I = np.eye(len(T_comp_t))

    Teff = np.linalg.inv(I - (T_comp_t @ C_sum_comp_t)) @ T_comp_t  # of cluster

    A_sca = Teff @ A_inc_comp_t

    data_m1[:, i] = A_sca.reshape(1, -1)

    print("iterations done:" + str(i + 1) + "/" + str(len(winc)))

print("generating figure...")


# generating figures using data


fig, ax = plt.subplots(4, 3, figsize=(7, 7), sharey="row", sharex="col")

for ud in [-1, 1]:
    for s in [0, 1]:
        if s == 0:
            al = [(0, 0), (1, 0), (1, 1)]
        else:
            al = [(0, 0), (0, 1), (1, 1)]

        for pq in al:
            if s == 0 and ud == -1:
                if pq == (0, 0):
                    ab = (0, 0)
                if pq == (1, 0):
                    ab = (0, 1)
                if pq == (1, 1):
                    ab = (0, 2)
            elif s == 1 and ud == -1:
                if pq == (0, 0):
                    ab = (1, 0)
                if pq == (0, 1):
                    ab = (1, 1)
                if pq == (1, 1):
                    ab = (1, 2)
            elif s == 0 and ud == 1:
                if pq == (0, 0):
                    ab = (2, 0)
                if pq == (1, 0):
                    ab = (2, 1)
                if pq == (1, 1):
                    ab = (2, 2)
            else:
                if pq == (0, 0):
                    ab = (3, 0)
                if pq == (0, 1):
                    ab = (3, 1)
                if pq == (1, 1):
                    ab = (3, 2)

            for j in (-jo - 1, -jo, -jo + 1):

                count = 0  # for seperating evancescent and propagating orders

                Ejpqs_coeff = np.zeros(len(winc))

                for i in range(0, len(winc)):

                    wj = winc[i] + (
                        (j + jo) * wm
                    )  # frequency at which the scattered field is calculated

                    ev = cmath.sqrt(
                        wj ** 2 - (pq[0] ** 2 + pq[1] ** 2) * w_l ** 2
                    )  # works only for perpendicular inc1

                    if not np.isreal(ev):
                        count += 1

                    Ejpqs_coeff[i] = np.abs(
                        get_Ejpqs_coeff(
                            kpar, wj, j, jo, pq, s, sinc, l_max, a, ud, data_m1[:, i]
                        )
                    )  # works only for perpendicular inc

                if j == -jo - 1:

                    col = "r"

                elif j == -jo:

                    col = "g"

                elif j == -jo + 1:

                    col = "b"

                ax[ab].text(
                    0.8, 1.05, "$\mathbf{g}^{\prime}_{" + str(pq[0]) + str(pq[1]) + "}$"
                )
                ax[ab].grid()

                if count == 0:
                    ax[ab].plot((winc / wn), Ejpqs_coeff, linewidth=1.5, color=col)

                else:
                    ax[ab].plot(
                        (winc / wn)[0:count],
                        Ejpqs_coeff[0:count],
                        "--",
                        dashes=(1.5, 1),
                        linewidth=1.5,
                        color=col,
                    )
                    ax[ab].plot(
                        (winc / wn)[count - 1 : -1],
                        Ejpqs_coeff[count - 1 : -1],
                        linewidth=1.5,
                        color=col,
                    )


ax[0, 2].text(1.1, 0.5, "$s^{\prime}=\mathrm{M}$")
ax[1, 2].text(1.1, 0.5, "$s^{\prime}=\mathrm{N}$")
ax[2, 2].text(1.1, 0.5, "$s^{\prime}=\mathrm{M}$")
ax[3, 2].text(1.1, 0.5, "$s^{\prime}=\mathrm{N}$")

custom_xlim = (min(winc / wn), max(winc / wn))
custom_ylim = (0, 1.318)
plt.setp(ax, xlim=custom_xlim, ylim=custom_ylim)
plt.xticks(fontsize=10)
plt.yticks(fontsize=10)
plt.locator_params(axis="x", nbins=3)
plt.locator_params(axis="y", nbins=3)


line_labels = ["$j^{\prime}=j-1$", "$j^{\prime}=j$", "$j^{\prime}=j+1$"]
fig.legend(
    labels=line_labels,  # The labels for each line
    loc="right",  # Position of legend
    borderaxespad=0.05,  # Small spacing around legend box.
    fancybox=True,
    bbox_to_anchor=(0.3, 0.15),
    fontsize=10,
)

fig.text(0, 0.89, r"$|\mathbf{\hat{S}}|$")
fig.text(0.85, 0.01, r"$\omega_\mathrm{inc}/\omega_n $")
fig.text(0.0, 0.7, "Reflection", rotation=90)
fig.text(0.0, 0.2, "Transmission", rotation=90)


ax = ax.flat
for n, axs in enumerate(ax):

    axs.text(
        0.01, 0.85, string.ascii_lowercase[n] + ")", transform=axs.transAxes, size=10
    )
plt.tight_layout()
plt.show()
