import numpy as np
from scipy import special as sps
from scipy.linalg import lu_factor, lu_solve

##spherical Bessel function of first kind (also holds for negative reals)

"""
---input
nu: order of spherical bessel function (scalar, integer)
z: domain point (float, complex) (cannot be an array)

---output
sp_J(nu,z): spehrical bessel function of first kind (float, complex) 

"""


def sp_J(nu, z):

    z = z.astype(complex)

    sph_J = sps.spherical_jn(nu, z, derivative=False)

    return sph_J


###Spherical Hankel function of first kind
"""
---input
nu: order of spherical bessel function (scalar, integer)
z: domain point (float, complex)

---output
sp_H(nu,z): spehrical hankel function of first kind (float, complex) (cannot be an array)

"""


def sp_H(nu, z):

    sph_J = sp_J(nu, z)  ##spherical bessel function of first kind

    sph_Y = sps.spherical_yn(
        nu, z, derivative=False
    )  ##spehrical bessel function of second kind

    sph_H = sph_J + 1j * sph_Y  ##spehrical hankel function of first kind

    return sph_H


"""
---input
nu: order of spherical bessel function (scalar, integer)
z: domain array (must always be a 1D numpy array)

---output
Zm1: 2D array of dimension len(z)xlen(z) 

"""


def Zm1(nu, z):

    Zm1 = np.zeros(len(z), dtype="complex")

    Zm1 = sp_J(nu, z)  ##spherical bessel array

    Zm1 = Zm1 * np.eye(len(Zm1))  ## converting array to matrix
    return Zm1


"""
for input, output, see Zm1
"""


def Zm3(nu, z):

    Zm3 = sp_H(nu, z)  ##spehrical hankel function of first kind

    Zm3 = Zm3 * np.eye(len(Zm3))  ## converting array to matrix

    return Zm3


"""
for input, output, see Zm1
"""


def Zn1(nu, z):

    a = sp_J(nu - 1, z)  ##spherical bessel funtion of first kind

    b = sp_J(nu + 1, z)  ##spherical bessel funtion of first kind

    Zn1 = a - nu / (2 * nu + 1) * (a + b)

    Zn1 = Zn1 * np.eye(len(Zn1))  ## converting array to matrix

    return Zn1


"""
for input, output, see Zm1
"""


def Zn3(nu, z):

    a = sp_H(nu - 1, z)  ##spherical hankel funtion of first kind

    b = sp_H(nu + 1, z)  ##spherical hankel funtion of first kind

    Zn3 = a - nu / (2 * nu + 1) * (a + b)

    Zn3 = Zn3 * np.eye(len(Zn3))  ## converting array to matrix

    return Zn3


##Kronceker delta function


def delta(a):
    if a == 0:

        return 1

    else:

        return 0


##associated legendre polynomial
"""
--input
m(integer, scalar): degree of legendre polynomial
l(integer, scalar): order of legendre polynomial
z(integer, scalar): argument of P_ml 
--output
P_ml(z) (complex, scalar): associated legendre polynomial at z

"""


def get_P_ml(m, l, z):

    a, b = sps.clpmn(
        m, l, z, type=2
    )  ## type=2 so that the brach cut is assumed at |z|>1

    return a[m, l]


##derivative of associated legendre polynomial
"""
--input
m(integer, scalar): degree of legendre polynomial
l(integer, scalar): order of legendre polynomial
z(integer, scalar): argument of P_ml 
--output
d_P_ml(z) (complex, scalar): associated legendre polynomial at z

"""


def get_d_P_ml(m, l, z):

    a, b = sps.clpmn(
        m, l, z, type=2
    )  ## type=2 so that the brach cut is assumed at |z|>1

    return b[m, l]


##calculating normalisation factor VSW as used in eq(22) of [2]
"""
--input 
l (integer, scalar ): Degree of the spherical wave
m (integer, scalar): Order of the spherical wave

--output
gamma_lm(scalar, complex): normalisation factor for VSW                
        
"""


def get_gamma_lm(l, m):

    return 1j * np.sqrt(
        ((2 * l + 1) * sps.factorial(l - m))
        / (4 * np.pi * l * (l + 1) * sps.factorial(l + m))
    )


##calculating area of a 2D unit cell

"""
--input
a (2x2 array): each row contains 1 lattice vector

--output
area: area of unit cell spanned by the vectors in a

"""


def get_area(a):

    return np.abs(np.cross(a[0], a[1]))


## calculating unit vector e_TE, TM(k_hat) for a given plane wave (characterised by k-vector)

"""
--input
k(3x1 array): contains kx, ky and kz components of k vector 
alpha(integer, scalar): 0 for TE and 1 for TM
 
--output
e_TE (1x3 array): TE unit vector  
e_TM (1x3 array): TM unit vector 
"""


def get_e_hat(alpha, k):

    kpar_n = np.sqrt(k[0] ** 2 + k[1] ** 2)

    k_norm = np.sqrt(k[0] ** 2 + k[1] ** 2 + k[2] ** 2)

    den = k_norm * kpar_n

    if alpha == 0:

        e_TE = np.zeros(3, dtype="complex")

        if kpar_n == 0 and k[2] >= 0:

            e_TE[0] = 0
            e_TE[1] = -1j
            e_TE[2] = 0

        elif kpar_n == 0 and k[2] < 0:

            e_TE[0] = 0
            e_TE[1] = -1j
            e_TE[2] = 0

        else:

            e_TE[0] = 1j * k[1] / kpar_n
            e_TE[1] = -1j * k[0] / kpar_n
            e_TE[2] = 0

        return e_TE

    if alpha == 1:

        e_TM = np.zeros(3, dtype="complex")

        if kpar_n == 0 and k[2] >= 0:

            e_TM[0] = -1
            e_TM[1] = 0
            e_TM[2] = 0

        elif kpar_n == 0 and k[2] < 0:

            e_TM[0] = 1
            e_TM[1] = 0
            e_TM[2] = 0

        else:

            e_TM[0] = -k[0] * k[2] / den
            e_TM[1] = -k[1] * k[2] / den
            e_TM[2] = kpar_n ** 2 / den

        return e_TM


## calculating theta for a given k-vector
"""
--input
kx(float): x- component of k vector 
ky(float): y- component of k vector 
kz(float): z- component of k vector 
--output
theta (complex, scalar): theta (spehrical coordinate) of given k-vector (range: [0,pi])
"""


def get_theta(kx, ky, kz):

    k_norm = np.sqrt(kx ** 2 + ky ** 2 + kz ** 2)

    return np.arccos(kz / k_norm)


## calculating phi for a given k-vector
"""
--input

kx(real): x- component of k vector 
ky(real): y- component of k vector 

--output
phi (complex, scalar): theta (spehrical coordinate) of given k-vector (range: [0,2pi])
"""


def get_phi(kx, ky):

    a = np.arctan2(ky, kx)

    if a >= 0:
        return a
    else:
        return 2 * np.pi + a


"""
---input

t(1D array, real): contains the time domain points
E_t (1D array, complex): time domain function 

---output
E_w (1D array, complex):frequency domain function

"""


def get_FT(t, E_t):

    dt = t[1] - t[0]

    E_w = len(t) * dt * np.fft.ifft(np.fft.ifftshift(E_t)) / np.sqrt(2 * np.pi)

    w = 2 * np.pi * (len(t) / np.max(t)) * np.fft.fftfreq(len(E_w))

    w = np.fft.fftshift(w)

    E_w = np.fft.fftshift(E_w)

    return w, E_w


"""
---input

w(1D array, real): contains the frequency domain points (should be even)

E_w (1D array, complex): frequency domain function 

---output
t((1D array, real): time array
  
E_t (1D array, real): time domain function

"""


def get_IFT(w, E_w):

    Fs = (w[1] - w[0]) / (2 * np.pi)

    E_w = np.fft.ifftshift(E_w)

    E_t = Fs * np.sqrt(2 * np.pi) * np.fft.fft(E_w)

    t = np.linspace(0, 1 / Fs, len(E_w))

    #print(t[2] - t[1])
    # print(t)

    return t, E_t


"""
---input

w(real): angular frequency

---output
lambda: wavelength

"""


def get_lmd(w):
    c_o = 3 * 1e8
    return (2 * np.pi * c_o) / w


"""
---input

lmd: wavelength

---output
w(real): angular frequency

"""


def get_w(lmd):
    c_o = 3 * 1e8
    return (2 * np.pi * c_o) / lmd


"""
---input
l_max(integer): maximum considered multipolar order

---output
L,M,ALPHA (1D arrays): modes

"""


def get_modes(l_max):
    dim = 2 * l_max * (l_max + 2)

    L = np.zeros(dim, dtype="int_")
    M = np.zeros(dim, dtype="int_")
    ALPHA = np.zeros(dim, dtype="int_")

    i = 0  # initialiser for index of L, M, ALPHA

    for alpha in range(0, 2):
        for l in range(1, l_max + 1):
            for m in range(-l, l + 1):
                L[i] = l
                M[i] = m
                ALPHA[i] = alpha
                i += 1

    return L, M, ALPHA


"""
---input
l_max(integer): maximum considered multipolar order
NOm(integer): number of frequencies in the comb

---output
W, L, M, ALPHA (1D arrays): modes

"""


def get_modes_t(lmax, NOm):

    dim = NOm * 2 * lmax * (lmax + 2)

    W = np.zeros(dim, dtype="int_")
    L = np.zeros(dim, dtype="int_")
    M = np.zeros(dim, dtype="int_")
    ALPHA = np.zeros(dim, dtype="int_")

    i = 0  # initialiser for index of L, M, ALPHA

    for w in range(0, NOm):

        for alpha in range(0, 2):

            for l in range(1, lmax + 1):

                for m in range(-l, l + 1):

                    W[i] = w
                    L[i] = l
                    M[i] = m
                    ALPHA[i] = alpha
                    i += 1

    return np.vstack([W, ALPHA, L, M])


##frequency domain gaussian pulse
"""
---input

w(real): doamin frequency

wo(real): centering frequency of the pulse
 
to(real): time delay of the pulse

To(real): duration of the pulse

Eo(complex): time domain amplitude


---output
A_w (1D array):frequency domain amplitude

"""


def get_gaussian(w, wo, to, To, Eo):

    return (
        -(
            Eo
            * To
            * (
                np.exp(-(To ** 2) * (w + wo) ** 2 / 2)
                + np.exp(-(To ** 2) * (w - wo) ** 2 / 2)
            )
            * np.exp(1j * w * to)
        )
        / 2
    )


"""
---input

w(real): doamin frequency

wo(real): centering frequency of the pulse
 
to(real): time delay of the pulse

To(real): duration of the pulse

Eo(complex): time domain amplitude


---output
E_inc_w (1D array):frequency domain amplitude of incident pulse

"""


def get_E_inc_wj(w, wo, to, To, Eo, polpw, kinc, r):

    E_inc_w = (
        get_gaussian(w, wo, to, To, Eo)
        * get_e_hat(polpw, kinc)
        * np.exp(1j * (kinc[2] * r[2]))
    )

    return E_inc_w


##creating a function to invert matrix without linalg.inv
"""
---input
A: matrix A
B: matrix B

---output
C: A^-1@B ,output result
"""


def solve(A, B):

    C = np.zeros(len(A) * len(A), dtype="complex").reshape(len(A), len(A))
    lu, piv = lu_factor(A)

    for i in range(0, len(B)):

        C[:, i] = lu_solve((lu, piv), B[:, i])

    return C


   
def get_conditioned_matrix(A):
    """Creating a function to condition a matrix
    Parameters
    ----------
    A : np.ndarray

    Returns
    -------
    A_new:np.ndarray
          conditioned matrix
    C: np.ndarray
       column conditioner
         
    """
    R= np.zeros((len(A),len(A)))
    C= np.zeros((len(A),len(A)))
                      
    for i in range (0, len(A)): #filling the column matrix
        R[i,i]=1/(np.sum(np.abs(A[i]))/len(A))
        C[i,i]=1/(np.sum(np.abs(A[:,i]))/len(A))
    
    A_new= R@A@C
        
    return A_new, C


#notes to improve

#fix negative frequencies in q-matrix
#fix the noise while doing frequency sweeps in file get_R
#why is BS folding when Ms=0???