import numpy as np
import scipy
import cmath
from ptsa import pw
from ptsa import sw
from ptsa.special._ufuncs import pi_fun, tau_fun
from ptsa.special._gufuncs import vsw_N, vsw_M
from scipy.sparse import block_diag
from sphere_functions import get_T_comp_t_element, S
from math_functions import (
    get_gamma_lm,
    get_area,
    get_theta,
    get_phi,
    get_e_hat,
    delta,
    get_modes
)


c_o = 3 * 1e8  # speed of light

# function to generate the incident field coefficients for the sphere at origion of a plane wave for single freqeuncy
"""
--input
modes(arrays): L,M, ALPHA modes 
ki_x(real): x-component of incident wave vector 
ki_y(real): y-component of incident wave vector 
ki_z(complex): z-component of incident wave vector 
polpw: Polarization of the plane wave (0 for TE and 1 for TM)
l_max: maximum order of spherical wave(integer & scalar >=1)

--output
A_inc: --1D array containing elements of incident field (dim: 1x\sum_{l=1}^{l_max}{2(2l+1)})
       -- components are ordered as: alpha(l(m))

"""


def get_A_inc_n(modes, ki_x, ki_y, ki_z, polpw, l_max):

    L, M, ALPHA = modes

    A_in = pw.to_sw(
        L, M, ALPHA, ki_x, ki_y, ki_z, polpw, posout=0, posin=0, helicity=False,
    )

    # returns incident field array corresponding to the given plane wave (for scatterer at the origin)
    return A_in


def get_A_inc(k, ki_x, ki_y, ki_z, polpw, l_max):

    a = 2 * l_max * (l_max + 2)

    A_in = np.zeros(a, dtype="complex")

    if k >= 0:

        i = 0
        for alpha in range(0, 2):
            for l in range(1, l_max + 1):
                for m in range(-l, l + 1):

                    A_in[i] = pw.to_sw(
                        l,
                        m,
                        alpha,
                        ki_x,
                        ki_y,
                        ki_z,
                        polpw,
                        posout=0,
                        posin=0,
                        helicity=False,
                    )
                    i += 1
    else:

        i = 0

        for alpha in range(0, 2):
            for l in range(1, l_max + 1):
                for m in range(-l, l + 1):
                    A_in[i] = (-1) ** (l + m + delta(alpha - 1)) * np.conjugate(
                        pw.to_sw(
                            l,
                            -m,
                            alpha,
                            ki_x,
                            ki_y,
                            -np.conjugate(ki_z),
                            polpw,
                            posout=0,
                            posin=0,
                            helicity=False,
                        )
                    )
                    i += 1

    # returns incident field array corresponding to the given plane wave (for scatterer at the origin)
    return A_in


# function to generate incident field coefficients for monochromatic input for time vaying media
"""
--input

j_o: appropriately chosen scalar for floquet comb (integer & scalar) (incident frequency acts as floquet frequency)
w_m: modulation frequency of the medium (float and scalar)
N_Om: total number of frequencies in the floquet comb
A_inc: single frequency column vector (1D array of dimension \sum_{l=1}^{l_max}{2(2l+1)})

--output
A_inc_t: incident field coefficient vector for time varying media (1D vector with dimension 1xN_Om*\sum_{l=1}^{l_max}{2(2l+1)})

"""


def get_A_inc_comp_t(j_o, N_Om, A_inc):

    A_in_t = np.zeros(0, dtype="complex")  # defining a null array to be filled

    A_o = np.zeros(len(A_inc))  # defining an array with zeros

    for j in range(0, N_Om):

        if (
            j == -j_o
        ):  # only entry corresponding to incident frequency yields non-zero contribution

            A_in_t = np.append(A_in_t, A_inc)

        else:

            A_in_t = np.append(A_in_t, A_o)

    A_in_t = np.array([A_in_t])

    A_in_t = A_in_t.transpose()  # converting row to a column vector

    return A_in_t


def get_T_comp_t(l_max, N_Om, k, S_k, k_o_ar, R):

    # calculating dimension of T_comp_t

    a = 2 * l_max * (l_max + 2)  #  length of A_inc

    dim = int(N_Om * a)  # so that the dimension of T_comp_t is dimxdim

    T_comp_t = np.zeros(dim * dim, dtype="complex").reshape(
        dim, dim
    )  # initialising T_comp_t

    # developing T_comp_t

    i = 0  # initialiser for row number of T_comp_t
    j = 0  # initialiser for column number of T_comp_t

    for w_o in range(0, N_Om):
        for alpha_o in range(0, 2):
            for l_o in range(1, l_max + 1):
                for m_o in range(-l_o, l_o + 1):
                    for w_s in range(0, N_Om):
                        for alpha_s in range(0, 2):
                            for l_s in range(1, l_max + 1):
                                for m_s in range(-l_s, l_s + 1):

                                    T_comp_t[i, j] = get_T_comp_t_element(
                                        w_o,
                                        l_o,
                                        m_o,
                                        alpha_o,
                                        w_s,
                                        l_s,
                                        m_s,
                                        alpha_s,
                                        k,
                                        S_k,
                                        k_o_ar,
                                        R,
                                    )
                                    j += 1

                    i += 1
                    j = 0

    return T_comp_t

def get_T_comp_t_n(l_max, N_Om, k, S_k, k_o_ar, R):

    # calculating dimension of T_comp_t

    a = 2 * l_max * (l_max + 2)  #  length of A_inc

    dim = int(N_Om * a)  # so that the dimension of T_comp_t is dimxdim

    T_comp_t = np.zeros(dim * dim, dtype="complex").reshape(
        dim, dim
    )  # initialising T_comp_t

    # developing T_comp_t

    i = 0  # initialiser for row number of T_comp_t
    j = 0  # initialiser for column number of T_comp_t
    ar= np.zeros(N_Om*2, dtype= int).reshape(2, N_Om)
    
    for w_o in range(0, N_Om):
        for alpha_o in range(0, 2):
            for l_o in range(1, l_max + 1):
                for m_o in range(-l_o, l_o + 1):
                    for w_s in range(0, N_Om):
                        for alpha_s in range(0, 2):
                            for l_s in range(1, l_max + 1):
                                for m_s in range(-l_s, l_s + 1):

                                    if m_s==-l_s and m_o==-l_o and l_o == l_s and alpha_o ==alpha_s and m_o ==m_s:
                                        
                                        T_comp_t[i, j] = get_T_comp_t_element(
                                        w_o,
                                        l_o,
                                        m_o,
                                        alpha_o,
                                        w_s,
                                        l_s,
                                        m_s,
                                        alpha_s,
                                        k,
                                        S_k,
                                        k_o_ar,
                                        R
                                    )
                                        ar[0,w_s]=i
                                        ar[1,w_s]=j
                                        
                                    elif l_o != l_s or alpha_o != alpha_s or m_o != m_s:
                                        
                                        T_comp_t[i, j] = get_T_comp_t_element(
                                        w_o,
                                        l_o,
                                        m_o,
                                        alpha_o,
                                        w_s,
                                        l_s,
                                        m_s,
                                        alpha_s,
                                        k,
                                        S_k,
                                        k_o_ar,
                                        R,
                                    )
                                    else:
                                        a= ar[0,w_s]
                                        b= ar[1,w_s]
                            
                                        T_comp_t[i, j]=T_comp_t[a,b]
                                    j += 1

                    i += 1
                    j = 0

    return T_comp_t


# creating summed up C matrix for single frequency
"""
--input

ks(float, scalar): angular wavenumber of considered scattered frequency 
kpar (float, 1x2 array): Parallel component of the incident wave, contains x and y component of \vec{k_{parallel}(\omega_o)}
a (float, 2x2 array): each row contains the lattice vectors
rs(float, 1x3 array): shift vector(?)
l_max(integer & scalar >=1): maximum order of spherical wave

--output
C_sum_ks : summed up C matrix for given frequency (characterised by angular wavenumber ks) and given incident wave (characterised by kpar)

"""


# caution: works only for real ks, because of eta


def get_C_sum_ks(modes, ks, kpar, a, rs):
    
    kref = 2 * 1e7  # reference wavenumber
    eta = 0 if np.abs(ks) <= kref else 0

    if ks < 0:
        modes = (modes[0], -modes[1], modes[2])
        ks = -ks

    C_sum_ks = sw.translate_periodic(
        ks, kpar, a, rs, modes, in_=None, rsin=None, helicity=False, eta=eta
    )
    return C_sum_ks



# creating summed up C matrix for all frequencies (block diagonal matrix) and monochromatic incidence


"""
--input
modes: L,M, ALPHA modes
j_o: appropriately chosen scalar for floquet comb (integer & scalar) (incident frequency acts as floquet frequency)
w_m: modulation frequency of the medium (float and scalar)
N_Om: total number of frequencies in the floquet comb
Omega(scalar): Floquet frequency
kpar (float, 1x2 array): Parallel component of the incident wave, contains x and y component of \vec{k_{parallel}(\omega_o)}
a (float, 2x2 array): each row contains the lattice vectors
rs(float, 1x3 array): shift vector(?)
l_max(integer & scalar >=1): maximum order of spherical wave
    
--output
C_sum_comp_t: summed up C matrix for all frequencies (block diagonal matrix) and monochromatic incidence (dim: N_Om*\sum_{l=1}^{l_max}{2(2l+1)}xN_Om*\sum_{l=1}^{l_max}{2(2l+1)})
"""


def get_C_sum_comp_t(modes, j_o, w_m, N_Om, Omega, kpar, a, rs, l_max):

    ar_i = np.arange(0, N_Om, 1)  # array of integerd from 0 to Nom-1

    w = Omega + (ar_i + j_o) * w_m
   
    k = w / c_o

    for j in range(0, N_Om):

        if k[j] >= 0:

            C_tr = get_C_sum_ks(modes, k[j], kpar, a, rs)

        else:

            C_tr = get_C_sum_ks(modes, k[j], kpar, a, rs)

            i1 = 0

            j1 = 0

            for alpha_o in range(0, 2):
                for l_o in range(1, l_max + 1):
                    for m_o in range(-l_o, l_o + 1):
                        for alpha_s in range(0, 2):
                            for l_s in range(1, l_max + 1):
                                for m_s in range(-l_s, l_s + 1):

                                    C_tr[i1, j1] = (-1) ** (
                                        l_o
                                        - l_s
                                        + m_s
                                        - m_o
                                        + delta(1 - alpha_o)
                                        - delta(1 - alpha_s)
                                    ) * np.conjugate(C_tr[i1, j1])

                                    j1 += 1
                        i1 += 1
                        j1 = 0

        if j == 0:

            C_t_tr = C_tr

        else:

            C_t_tr = block_diag((C_t_tr, C_tr))

    return C_t_tr


# Creating effective T-matrix for array of infinite array of time-varying spheres (eq(17) of [2])


"""
--input:
kx (float & scalar): x-component of wave vector of inc plane wave 
ky (float & scalar): y-component of wave vector of inc plane wave (float & scalar)
R (float & positive scalar): translation distance b/w two spheres  
theta_R (float & in rad.): polar angle  of \vec{R}
phi_R (float & in rad.) : azimuthal angle (float & in rad.) of \vec{R}
T_comp_t: full T_matrix of a single time varying sphere (dim: N_Om*\sum_{l=1}^{l_max}{2(2l+1)}xN_Om*\sum_{l=1}^{l_max}{2(2l+1)})
C_sum_t:  full C-matrix of the system (block diagonal wrt frequency) multiplied with e^(ikpar.r)  (dim: N_Om*\sum_{l=1}^{l_max}{2(2l+1)}xN_Om*\sum_{l=1}^{l_max}{2(2l+1)})

--output:

T_eff_t: T-matrix of time varying spherical cluster (dim: N_Om*\sum_{l=1}^{l_max}{2(2l+1)}xN_Om*\sum_{l=1}^{l_max}{2(2l+1)})
"""


def get_T_eff_t(T_comp_t, C_sum_comp_t):

    # defining identity matrix

    I = np.eye(len(T_comp_t))

    # creating matrix to be inverted later

    M = I - (T_comp_t @ C_sum_comp_t)

    # final T-matrix

    T_eff_t = np.linalg.inv(M) @ T_comp_t

    return T_eff_t


# creating the scattered field coefficients for specific frquency wj compatible with final lattice sum (eq. (??))
"""
--input

j(integer & scalar): index of wj
l_max(integer & scalar >=1): maximum order of spherical wave
theta_k (scalar, complex): theta angle of given scattered wave
A_sca(1xM array, complex): contains the scattered field vector of all frequencies  (M: NOm*\sum_{l=1}^{l_max}{2(2l+1)})

--output
A_sca_comp_t(1xM array, complex): contains the compatible (with PW sum) scattered field coefficients of given frequency  (M: \sum_{l=1}^{l_max}{2(2l+1)})
A_sca_TE(1xM/2 array, complex): contains the compatible TE (with PW sum) scattered field coefficients of given frequency  (M: \sum_{l=1}^{l_max}{2(2l+1)})
A_sca_TM(1xM/2 array, complex): contains the compatible TM(with PW sum) scattered field coefficients of given frequency  (M: \sum_{l=1}^{l_max}{2(2l+1)})

"""


def get_A_sca_comp_tem(j, l_max, theta_k, A_sca):

    # calculating dimension of A_sca_comp_t
    dim = 2 * l_max * (l_max + 2)

    dim2 = int(dim / 2)  # dimension for A_sca_TE,TM

    A_sca_comp_TE = np.zeros(dim2, dtype="complex")  # PW TE coefficients

    A_sca_comp_TM = np.zeros(dim2, dtype="complex")  # PW TM coefficients

    A_sca_TE = A_sca[dim * j : dim * j + dim2]  # filling VSW TM compoents

    A_sca_TM = A_sca[dim * j + dim2 : dim * j + 2 * dim2]  # filling VSW TM compoents

    # calculating PW scattered coefficients from VSW coefficients
    i = 0

    for l in range(1, l_max + 1):

        for m in range(-l, l + 1):

            A_sca_comp_TE[i] = (A_sca_TE[i] * tau_fun(l, m, np.cos(theta_k))) + (
                A_sca_TM[i] * pi_fun(l, m, np.cos(theta_k))
            )

            A_sca_comp_TM[i] = (A_sca_TE[i] * pi_fun(l, m, np.cos(theta_k))) + (
                A_sca_TM[i] * tau_fun(l, m, np.cos(theta_k))
            )

            i += 1

    A_sca_comp_t = np.hstack([A_sca_comp_TE, A_sca_comp_TM])

    return A_sca_comp_t, A_sca_comp_TE, A_sca_comp_TM


# creating function to generate the scattered field of a bi-periodoc array of infinite spheres

"""
--input
kpar(1x2 array): array containing parallel components of the incident wave vector
wj(scalar): scattered wave frequency
j(integer): index of wj
l_max(integer & scalar >=1): maximum order of spherical wave
a (2x2 array): each row contains 1 lattice vector (only works in a[0] is in x-direction and a[1] is in y-direction)
r(1x3 array): position vector in space at which field is calculated
A_sca(1xM array, complex): contains the scattered field vector of all frequencies  (M: NOm*\sum_{l=1}^{l_max}{2(2l+1)})

--output
E_sca_wj(1x3 array, complex): contains the scattered field vector of given frequency  

caution: works only for real k_ix, k_iy (because of arctan2)

"""


def get_E_sca_wj(kpar, wj, j, l_max, a, r, A_sca):

    E_sca_wj = np.zeros(3, dtype="complex")  # initialising scattered field at wj

    kj = wj / c_o  # angular wavenumber corresponding wj

    g = np.zeros(2)  # \vec{g}

    for p in range(-4, 5):  # sum of \vec{g}

        for q in range(-4, 5):

            g[0] = 2 * np.pi * p / (np.linalg.norm(a[0]))

            g[1] = 2 * np.pi * q / (np.linalg.norm(a[1]))

            kg_par = kpar + g

            Gamma_k_g = cmath.sqrt(kj ** 2 - kg_par[0] ** 2 - kg_par[1] ** 2)

            if r[2] >= 0:

                kg_z = Gamma_k_g

            else:

                kg_z = -Gamma_k_g

            k_g = np.append(kg_par, kg_z)

            if kj < 0:

                k_g[2] = np.conjugate(k_g[2])

                Gamma_k_g = np.conjugate(Gamma_k_g)

            phi_k = get_phi(kg_par[0], kg_par[1])

            theta_k = get_theta(k_g[0], k_g[1], k_g[2])

            A_sca_comp_t, A_TE, A_TM = get_A_sca_comp_tem(j, l_max, theta_k, A_sca)

            i = 0

            for alpha in range(0, 2):

                for l in range(1, l_max + 1):

                    for m in range(-l, l + 1):

                        if kj > 0:

                            E_sca_wj += (
                                2
                                * np.pi
                                * get_gamma_lm(l, m)
                                * np.exp(1j * m * phi_k)
                                * np.exp(1j * np.dot(k_g, r))
                                * A_sca_comp_t[i]
                                * get_e_hat(alpha, k_g)
                            ) / (1j ** (l + 1) * get_area(a) * kj * Gamma_k_g)

                        else:

                            E_sca_wj += (
                                2
                                * np.pi
                                * get_gamma_lm(l, m)
                                * np.exp(1j * m * phi_k)
                                * np.exp(-1j * np.dot(k_g, r))
                                * A_sca_comp_t[i]
                                * get_e_hat(alpha, k_g)
                            ) / (1j ** (l + 1) * get_area(a) * kj * Gamma_k_g)

                        i += 1

    return E_sca_wj


def get_d_E_sca_wj(kpar, wj, j, l_max, a, r, A_sca):

    c_o = 3 * 1e8

    kj = wj / c_o

    E_sca = np.zeros(3, dtype="complex")

    R = np.zeros(3)

    dim = 2 * l_max * (l_max + 2)

    dim2 = int(dim / 2)

    A_sca_j = A_sca[dim * j : dim * (j + 1)]

    for p in range(-50, 50):
        for q in range(-50, 50):

            R[0] = p * a[0, 0]
            R[1] = q * a[1, 1]

            if kj >= 0:

                k_rR = kj * np.linalg.norm(r - R)

                rR = r - R

                theta = get_theta(rR[0], rR[1], rR[2])

                phi = get_phi(rR[0], rR[1])

                i = 0

                for l in range(1, l_max + 1):

                    for m in range(-l, l + 1):

                        E_sca += (A_sca_j[i] * vsw_M(l, m, k_rR, theta, phi)) + (
                            A_sca_j[i + dim2] * vsw_N(l, m, k_rR, theta, phi)
                        )

                        i += 1

            else:

                k_rR = -kj * np.linalg.norm(r - R)

                rR = r - R

                theta = get_theta(rR[0], rR[1], rR[2])

                phi = get_phi(rR[0], rR[1])

                i = 0

                for l in range(1, l_max + 1):

                    for m in range(-l, l + 1):

                        E_sca += (
                            A_sca_j[i]
                            * (-1) ** (l + m)
                            * np.conjugate(vsw_M(l, -m, k_rR, theta, phi))
                        ) + (
                            A_sca_j[i + dim2]
                            * (-1) ** (l + m + 1)
                            * np.conjugate(vsw_N(l, -m, k_rR, theta, phi))
                        )

                        i += 1

    return E_sca


# time averaged poynting vector normalised by area of field given in PW parity basis
#for a given frequency wj in the comb

"""
--input
kpar(1x2 array): array containing parallel components of the incident wave vector
wj(scalar): scattered wave frequency
j(int): index of wj
kinc(int): wavenumber of inc wave
lmax(int >=1): maximum order of spherical wave
a (2x2 array): each row contains 1 lattice vector (only works in a[0] is in x-direction and a[1] is in y-direction)
ud(1x3 array): ud=1 for T and ud=-1 for R
Asca(1xM array, complex): contains the scattered field vector of all frequencies  (M: NOm*\sum_{l=1}^{l_max}{2(2l+1)})
Egsd_i(1x2 array, complex): incident field components in PW parity basis of frequency jinc

--output
TR(real): Transmittivity or Reflectivity, respectively

caution: works only for real k_ix, ki_y (because of arctan2) and for positive frequencies and air as embedding medium and upward propapgating
         monochromatic inc. wave

"""


def get_RT(kpar, wj, j, kinc, lmax, a, ud, Asca, Egsd_i):

    kj = wj / c_o  # angular wavenumber corresponding wj

    g = np.zeros(2)  # \vec{g}

    ngmax = 2  # max no of diffraction orders considered

    Egsd = np.zeros(
        2 * (2 * ngmax + 1) ** 2, dtype="complex"
    )  # initialising scattered field coefficents

    i = 0

    for p in range(-ngmax, ngmax + 1):  # sum of \vec{g}

        for q in range(-ngmax, ngmax + 1):

            g[0] = 2 * np.pi * p / (np.linalg.norm(a[0]))

            g[1] = 2 * np.pi * q / (np.linalg.norm(a[1]))

            kg_par = kpar + g

            Gamma_k_g = cmath.sqrt(kj ** 2 - kg_par[0] ** 2 - kg_par[1] ** 2)

            if ud >= 0:

                kg_z = Gamma_k_g

            else:

                kg_z = -Gamma_k_g

            k_g = np.append(kg_par, kg_z)

            phi_k = get_phi(kg_par[0], kg_par[1])

            theta_k = get_theta(k_g[0], k_g[1], k_g[2])

            A_sca_comp_t, A_TE, A_TM = get_A_sca_comp_tem(j, lmax, theta_k, Asca)

            b = 0

            for alpha in range(0, 2):

                for l in range(1, lmax + 1):

                    for m in range(-l, l + 1):

                        Egsd[i] += (
                            2
                            * np.pi
                            * get_gamma_lm(l, m)
                            * np.exp(1j * m * phi_k)
                            * A_sca_comp_t[b]
                        ) / (1j ** (l + 1) * get_area(a) * kj * Gamma_k_g)

                        b += 1

                if p == 0 and q == 0 and ud == 1 and kj == kinc:  # adding inc field

                    if alpha == 0:
                        Egsd[i] += Egsd_i[0]
                    elif alpha == 1:
                        Egsd[i] += Egsd_i[1]

                i += 1

    # calculating for output field
    Sav_o = 0

    i = 0

    for p in range(-ngmax, ngmax + 1):

        for q in range(-ngmax, ngmax + 1):

            g[0] = 2 * np.pi * p / (np.linalg.norm(a[0]))

            g[1] = 2 * np.pi * q / (np.linalg.norm(a[1]))

            kg_par = kpar + g

            Gamma_k_g = cmath.sqrt(kj ** 2 - kg_par[0] ** 2 - kg_par[1] ** 2)

            if ud >= 0: #for transmissivity

                kg_z = Gamma_k_g

            else: #for reflectivity

                kg_z = -Gamma_k_g

            Sav_o += (1 / kj) * np.real(
                kg_z * np.abs(Egsd[i + 1]) ** 2
                + np.conjugate(kg_z) * np.abs(Egsd[i]) ** 2
            )

            i += 2

    # for incident field

    Sav_i = 0

    for i in range(0, 2):

        kz = np.sqrt(kinc ** 2 - kpar[0] ** 2 - kpar[1] ** 2)

        if i == 0:

            Sav_i += (1 / kinc) * np.abs(Egsd_i[i]) ** 2 * np.real(np.conjugate(kz))

        else:

            Sav_i += (1 / kinc) * np.abs(Egsd_i[i]) ** 2 * np.real(kz)

    return Sav_o / Sav_i
    


# Band structure for a 4D spatiotemporal crystal

"""
--input
G_ar(1xM vector): array containing wavevetors in the first BZ
Om_ar(1xM vector): array containing Floquet frequecies
lmax(integer & scalar >=1): maximum order of spherical wave
jo (int): appropriate integer such that assigned values of j span the truncated spectrum (scalar)
NOm (int): number of frequencies in 1 periodic comb of the truncated spectrum (scalar, integer)
Ms (scalar): initial modulation strength 
wm (scalar): modulation frequency of the medium
wn (scalar): natural frequency of oscillator 
yn (scalar): damping factor of oscillator
No (int): unmodulated electron density
R(scalar): radius of sphere
a (3x3 array): matix containing lattice vectors, each row contains 1 lattice vector 
rs (1x3 vector): shift vector

--output
band_structure: matrix containing condition numbers of the matrix (I-TC) with frequency on x-axis and wavevector on y-axis
"""


def get_band_structure(G_ar, Om_ar, lmax, jo, NOm, Ms, wm, wn, yn, No, R, a, rs):

    itera = 0  # counter to track iterations
    modes = get_modes(lmax)
    band_structure = np.zeros(len(G_ar) * len(Om_ar)).reshape(
        len(Om_ar), len(G_ar)
    )  # initialising the band structure matrix

    for p in range(0, 6):

        cond = np.zeros(len(G_ar) * len(Om_ar)).reshape(
            len(Om_ar), len(G_ar)
        )  # initialising the condition number matrix

        for i in range(0, len(Om_ar)):

            for j in range(0, len(G_ar)):

                if p == 0:
                    kpar = np.array([0, G_ar[j], 0])

                elif p == 1:
                    kpar = np.array([G_ar[j], np.pi / a[0, 0], 0])

                elif p == 2:
                    kpar = np.array(
                        [G_ar[len(G_ar) - j - 1], G_ar[len(G_ar) - j - 1], 0]
                    )

                elif p == 3:
                    
                    kpar = np.array([G_ar[j], G_ar[j], G_ar[j]])

                elif p == 4:
                    kpar = np.array(
                        [
                            G_ar[len(G_ar) - j - 1],
                            np.pi / a[0, 0],
                            G_ar[len(G_ar) - j - 1],
                        ]
                    )

                elif p == 5:
                    kpar = np.array([np.pi / a[0, 0], np.pi / a[0, 0], G_ar[j]])

                k, S_k, k_o_ar = S(Om_ar[i], NOm, jo, No, wn, yn, Ms, wm)

                T_comp_t = get_T_comp_t_n(
                    lmax, NOm, k, S_k, k_o_ar, R
                )  # T_mat of single sphere
                
                C_sum_comp_t = get_C_sum_comp_t(
                    modes, jo, wm, NOm, Om_ar[i], kpar, a, rs, lmax
                )  # translation matrix

                I = np.eye(len(T_comp_t))
                
                M= I-T_comp_t@C_sum_comp_t
                
                M=scipy.linalg.matrix_balance(M)[0]
                
                cond[i,j]= min(np.abs(scipy.linalg.svdvals(M)))
                
                #cond[i,j]=np.abs(np.linalg.det(scipy.linalg.matrix_balance(M)[0]))
                
                itera += 1  # appending iterations
                print(
                    "iterations done: "
                    + str(itera)
                    + "/"
                    + str(len(Om_ar) * 6*len(G_ar))
                )

        band_structure = np.append(band_structure, cond, axis=1)

    return band_structure



# creating function to generate the trasmitted and reflected field of given spectral and spatial diffraction order and parity
"""
--input
kpar(1x2 array): array containing parallel components of the incident wave vector
wj(scalar): scattered wave frequency
j(int): index of wj
jo(int): approprately chosen integer
pq(tuple, int): containing indices pq of considered diffraction order g_pq
l_max(integer & scalar >=1): maximum order of spherical wave
a (2x2 array): each row contains 1 lattice vector (only works in a[0] is in x-direction and a[1] is in y-direction)
s(int): 0 for TE and 1 for TMs for scattered wave
sinc(int): 0 for TE and 1 for TM for incidence wave
ud(int): +1 for transmitted and -1 for reflected
A_sca(1xM array, complex): contains the scattered field vector of all frequencies  (M: NOm*\sum_{l=1}^{l_max}{2(2l+1)})

--output
Ejpqs_coeff(1x3 array, complex): contains the scattered field vector of given frequency  

caution: works only for real k_ix, k_iy (because of arctan2) and if w_-jo= winc and normal inc

"""


def get_Ejpqs_coeff(kpar, wj, j, jo, pq, s, sinc, l_max, a, ud, A_sca):

    Ejpqs_coeff = 0  # initialising scattered coefficient at wj

    kj = wj / c_o  # angular wavenumber corresponding wj

    g = np.zeros(2)  # \vec{g}

    g[0] = 2 * np.pi * pq[0] / (np.linalg.norm(a[0]))

    g[1] = 2 * np.pi * pq[1] / (np.linalg.norm(a[1]))

    if kj >= 0:

        kg_par = kpar + g

    else:

        kg_par = -kpar + g

    Gamma_k_g = cmath.sqrt(kj ** 2 - kg_par[0] ** 2 - kg_par[1] ** 2)

    if ud == 1:

        kg_z = Gamma_k_g

    else:

        kg_z = -Gamma_k_g

    k_g = np.append(kg_par, kg_z)

    if kj < 0:

        k_g = np.conjugate(k_g)

        Gamma_k_g = np.conjugate(Gamma_k_g)

    phi_k = get_phi(kg_par[0], kg_par[1])

    theta_k = get_theta(k_g[0], k_g[1], k_g[2])

    A_sca_comp_t, A_TE, A_TM = get_A_sca_comp_tem(j, l_max, theta_k, A_sca)
    
    if s== 0:
        
        A= A_TE
        
    elif s==1:
        
        A= A_TM
        
    i = 0
    

    for l in range(1, l_max + 1):
        for m in range(-l, l + 1):
                
            if kj > 0:
                    
                Ejpqs_coeff+= (  2
                                    * np.pi
                                    * get_gamma_lm(l, m)
                                    * np.exp(1j * m * phi_k)
                                    * A[i]
                                ) / (1j ** (l + 1) * get_area(a) * kj * Gamma_k_g)
            else:
                Ejpqs_coeff+= (    2
                                    * np.pi
                                    * get_gamma_lm(l, m)
                                    * np.exp(1j * m * phi_k)
                                    * A[i]
                                ) / (1j ** (l + 1) * get_area(a) * kj * Gamma_k_g)
    
            i += 1
            
    if pq[0]== 0 and pq[1]== 0 and ud== 1 and s== sinc and j== -jo:
        
        Ejpqs_coeff+=1
        
    return Ejpqs_coeff

