import numpy as np
from math_functions import delta
from math_functions import Zm1, Zm3, Zn1, Zn3
from scipy.linalg import matrix_balance

##natural constants in SI units

e = 1.6 * 1e-19  ##charge on an electron

m_e = 9.11 * 1e-31  ##mass of an electron

eps_o = 8.854 * 1e-12  ##free space permittivity

c_o = 3 * 1e8  ##speed of light in vaccum


###Defining response function in frequency domain (implements eq. (18) of [1])

"""
--input
w1,w2: frequency arguments (scalars)
N_o: unmodulated electron density (scalar, integer)
w_n= natural frequency of oscillator (scalar)
y_n= damping factor of oscillator (scalar)
M_s= initial modulation strength (scalar)
w_m= modulation frequency of the medium (scalar)
--output
R_end: response function of a time varying non-dispersive material (scalar)
"""


def R_end(w1, w2, N_o, w_n, y_n, M_s, w_m):

    return 11.5 * (delta(w1) + (M_s / 2.0) * (delta(w1 + w_m) + delta(w1 - w_m)))



"""
--input
w1,w2: frequency arguments (scalars)
N_o: unmodulated electron density (scalar, integer)
w_n= natural frequency of oscillator (scalar)
y_n= damping factor of oscillator (scalar)
M_s= initial modulation strength (scalar)
w_m= modulation frequency of the medium (scalar)
--output
R_e: response function of a time varying dispersive lorentzian material (scalar)
"""


def R_e(w1, w2, N_o, w_n, y_n, M_s, w_m):

    return (
        (N_o * e ** 2 / (m_e * eps_o))
        * (delta(w1) + (M_s / 2.0) * (delta(w1 + w_m) + delta(w1 - w_m)))
        / (w_n ** 2 - (w1 + w2) ** 2 - 1j * y_n * (w1 + w2))
    )


def N(
    w1, w2, N_o, M_s, w_m
):  ##FT of electron density to be compatible with matlab code

    return (
        np.sqrt(2 * np.pi)
        * N_o
        * (delta(w1) + (M_s / 2.0) * (delta(w1 + w_m) + delta(w1 - w_m)))
    )


##defining function that returns eigenvalues and vectors for each floquet frequency (implements eq. (8) of [1])
"""
---input

Omega= Floquet frequency (scalar)
N_Om= number of frequencies in 1 periodic comb of the truncated spectrum (scalar, integer)
j_o= appropriate integer such that assigned values of j span the truncated spectrum (scalar)
N_o= unmodulated electron density (scalar, integer)
w_n= natural frequency of oscillator (scalar)
y_n= damping factor of oscillator (scalar)
M_s= initial modulation strength (scalar)
w_m= modulation frequency of the medium (scalar)

---output
k: sqrt of eigenvalues (1D array of dim. 1xN_Om)
k2: eigenvalues of \hat{K} (1D array of dim. 1xN_Om)
S_k: eigenvectors of \hat{K} (2D array of dim. N_OmxN_Om, where, each column corresponds to an eigen vector)
k_o_ar: array of freespece wavenumbers corresponding to frequencies in the floquet comb ((1D array of dim. 1xN_Om))

"""


def S(Omega, N_Om, j_o, N_o, w_n, y_n, M_s, w_m):
    
    ## defining spectral comb for given Floquet frequency

    ar_i = np.arange(0, N_Om, 1)  # array of integers from 0 to NOm-1

    Om = Omega + (ar_i + j_o) * w_m

    ## defining the matrices

    ## identity matrix
    I = np.eye(N_Om)

    ## k_o matrix

    k_o = Om / c_o

    k_o_ar = k_o  ##saving k_o array for returning in the end

    k_o = I * k_o

    ## response function matrix

    R = np.zeros(N_Om * N_Om, dtype="complex").reshape(N_Om, N_Om)
    N0 = np.zeros(N_Om * N_Om, dtype="complex").reshape(
        N_Om, N_Om
    )  ## this line is added to be compatible with matlab code

    for j in range(0, N_Om):

        for l in range(0, N_Om):

            R[j, l] = R_e(Om[j] - Om[l], Om[l], N_o, w_n, y_n, M_s, w_m)
            
            N0[j, l] = N(Om[j] - Om[l], Om[l], N_o, M_s, w_m)
    
    ##total input matrix

    K = k_o @ k_o @ (I + R)

    ## calculating eigenvalues and vectors
    k2, S_k = np.linalg.eig(K)  ##computes eigenvalues and vectors of K

    idx = np.argsort(k2)  ##index according to ascending k2
    k2 = k2[idx]  ##sorting the eigenvalues
    S_k = S_k[:, idx]  ##sorting the eigenvectors
    k = np.sqrt(k2)  ##square root of eigenvalues

    return k, S_k, k_o_ar


# function for getting T-matrix of spherical scatterer

"""---input
k: array containing sqrt. of eigenvalues of S (1D array of dim. N_Omx1) 
S_k: eigen vectors of K (Matrix of dim. N_OmxN_Om with each column representing the eigenvector)
k_o_ar: array containing free space wavenumbers corresponding to frequencies of a floquet comb (1D array of dim. N_Omx1) 
R: radius of sphere (scalar)
l: multipolar order (order of bessel function) (scalar)
alpha: integer that assigns polarization of required VSW and takes values,
       '0' for TE polarization
       '1' for TM polarization    

---output

T_sca_t: scattering T-matrix of a dispersive time-varying sphere given by eq30 in [1] (Matrix of dimension N_OmxN_Om)

"""


def get_T_sca(k, S_k, k_o_ar, R, l, alpha):
    
    k_mat = k * np.eye(len(k))  ##returns diagonal k matrix
    
    if alpha == 0:

        ##A is first matrix of T_mat

        A11 = S_k @ Zm1(l, k * R)
        A12 = -Zm3(l, k_o_ar * R)
        A21 = np.eye(len(k_o_ar)) * (1 / k_o_ar) @ S_k @ k_mat @ Zn1(l, k * R)
        A22 = -Zn3(l, k_o_ar * R)

        ##B is second matrix of T_mat
        N_Om = len(k_o_ar)  ##dimension of zero matrix is N_OmxN_Om

        B11 = Zm1(l, k_o_ar * R)
        B12 = np.zeros(N_Om * N_Om).reshape(N_Om, N_Om)
        B21 = B12
        B22 = Zn1(l, k_o_ar * R)

    if alpha == 1:
        ##A is first matrix of T_mat

        A11 = S_k @ Zn1(l, k * R)
        A12 = -Zn3(l, k_o_ar * R)
        A21 = np.eye(len(k_o_ar)) * (1 / k_o_ar) @ S_k @ k_mat @ Zm1(l, k * R)
        A22 = -Zm3(l, k_o_ar * R)

        ##B is second matrix of T_mat
        N_Om = len(k_o_ar)  ##dimension of zero matrix is N_OmxN_Om

        B11 = Zn1(l, k_o_ar * R)
        B12 = np.zeros(N_Om * N_Om).reshape(N_Om, N_Om)
        B21 = B12
        B22 = Zm1(l, k_o_ar * R)

    A = np.block([[A11, A12], [A21, A22]])
    
    M, T = matrix_balance(A)
    
    Ainv=T@np.linalg.inv(M)@np.linalg.inv(T)  #inverse of A
    
    #A = np.linalg.inv(A)  ##inverting A as needed by eq (28)
    ##generating the second matrix of T_mat
    B = np.block([[B11, B12], [B21, B22]])

    ##final T matrix

    T = Ainv @ B

    ##extracting the scattered T-matrix from total T matrix

    T_sca_t = T[N_Om : 2 * N_Om, 0:N_Om] + T[N_Om : 2 * N_Om, N_Om : 2 * N_Om]
    
    
    return T_sca_t


## 8 argument function that returns 1 element of T_comp_t

"""
--input
w_o(scalar & integer): output frequency index (goes from 0 to N_Om-1)
l_o(scalar & integer): output order of VSW
m_o(scalar & integer): output degree of VSW
alpha_o(scalar & integer): output parity of VSW (0 for TE and 1 for TM)
w_s(scalar & integer): source frequency index (goes from 0 to N_Om-1)
l_s(scalar & integer): source order of VSW
m_s(scalar & integer): source degree of VSW
alpha_s(scalar &integer): sourceparity of VSW (0 for TE and 1 for TM)
get_T_sca(l, alpha): T_sca for given alpha and l 
k (1D array of dim. N_Omx1) : array containing sqrt. of eigenvalues of S
S_k (Matrix of dim. N_OmxN_Om): eigen vectors of K (each column representing the eigenvector)
k_o_ar (1D array of dim. N_Omx1): array containing free space wavenumbers corresponding to frequencies of a floquet comb 
R: radius of sphere (scalar)

--output
T_comp_t_element: 1 element of T_comp_t
"""


def get_T_comp_t_element_v(out_c, in_c, k, S_k, k_o_ar, R):

    if out_c[1] != in_c[1] or out_c[2] != in_c[2] or out_c[3] != in_c[3]:

        T_comp_t_element = 0

    else:

        T_comp_t_element = get_T_sca(k, S_k, k_o_ar, R, out_c[2], out_c[1])[
            out_c[0], in_c[0]
        ]

    return T_comp_t_element


## 8 argument function that returns 1 element of T_comp_t

"""
--input
w_o(scalar & integer): output frequency index (goes from 0 to N_Om-1)
l_o(scalar & integer): output order of VSW
m_o(scalar & integer): output degree of VSW
alpha_o(scalar & integer): output parity of VSW (0 for TE and 1 for TM)
w_s(scalar & integer): source frequency index (goes from 0 to N_Om-1)
l_s(scalar & integer): source order of VSW
m_s(scalar & integer): source degree of VSW
alpha_s(scalar &integer): sourceparity of VSW (0 for TE and 1 for TM)
get_T_sca(l, alpha): T_sca for given alpha and l 
k (1D array of dim. N_Omx1) : array containing sqrt. of eigenvalues of S
S_k (Matrix of dim. N_OmxN_Om): eigen vectors of K (each column representing the eigenvector)
k_o_ar (1D array of dim. N_Omx1): array containing free space wavenumbers corresponding to frequencies of a floquet comb 
R: radius of sphere (scalar)

--output
T_comp_t_element: 1 element of T_comp_t
"""


def get_T_comp_t_element(
    w_o, l_o, m_o, alpha_o, w_s, l_s, m_s, alpha_s, k, S_k, k_o_ar, R
): 
    
    if l_o != l_s or alpha_o != alpha_s or m_o != m_s:

        T_comp_t_element = 0
        
    else:
        
        T_comp_t_element = get_T_sca(k, S_k, k_o_ar, R, l_s, alpha_s)[w_o, w_s]

    return T_comp_t_element
